/*   _ants.c --- Ant Colony
 *
 *   Copyright (C) 2000 Valvassori Mose <djedi@altern.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA **/

#define _ANTS_C
#include <stdlib.h>
#include <math.h>
#include "ants.h"

/* This array is static 'cause, I can easily control
   the number of ants with the slider (see option.c) */
ANT ant[MAX_ANTS];						 /*+ The ant colony +*/
int nbants=1500;

int active_trail_start= _active_trail_start;
int active_trail_end= _active_trail_end;

char use_pheronom=1;
char use_noise=1;
float ant_noise_trace=0.3;
float ant_noise_attraction;
float ant_noise_go_home=.2;


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Initialize an ant.

  int i Ant's ID
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
init_ants(int i)
{
  ant_noise_attraction=ant_noise_trace/2;
  if (!from_nest)
    {
      ant[i].posx = ((float)MW)*rand()/(RAND_MAX+1.0);
      ant[i].posy = ((float)MH)*rand()/(RAND_MAX+1.0);
    }
  else
    {
      ant[i].posx = NESTX;
      ant[i].posy = NESTY;
    }
  ant[i].direction = 8.0*rand()/(RAND_MAX+1.0);
  ant[i].state= searching;
  ant[i].food=0;
}









/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  This ant use a random walk.
  Modify directly the "ants" array

  int i the ant number id.
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
random_walk(int i)
{
  /*  The 8 directions map:
   *
   *         1 0 7
   *          \|/
   *       2 --+-- 6
   *          /|\
   *         3 4 5
   */
  int direction;
  int r;

  r = 3.0*rand()/RAND_MAX;
  r -=1;

  direction = (ant[i].direction + r);
  direction = (direction + 8) % 8;

  switch (direction)
    {
    case 0:
      if (ant[i].posy>0)
	ant[i].posy --;
      break;
    case 1:
      if ((ant[i].posx>0) && (ant[i].posy>0))
	{
	  ant[i].posy --;
	  ant[i].posx --;
	}
      break;
    case 2:
      if (ant[i].posx>0)
	ant[i].posx --;
      break;
    case 3:
      if ((ant[i].posx>0) && (ant[i].posy<MH-1))
	{
	  ant[i].posy ++;
	  ant[i].posx --;
	}
      break;
    case 4:
      if (ant[i].posy<MH-1)
	ant[i].posy ++;
      break;
    case 5:
      if ((ant[i].posx<MW-1) && (ant[i].posy<MH-1))
	{
	  ant[i].posy ++;
	  ant[i].posx ++;
	}
      break;
    case 6:
      if (ant[i].posx<MW-1)
	ant[i].posx ++;
      break;
    case 7:
      if ((ant[i].posx<MW-1) && (ant[i].posy>0))
	{
	  ant[i].posy --;
	  ant[i].posx ++;
	}
      break;
    };
  ant[i].direction = direction;
}





/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  The pheronoms attract the ant.

  int i
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
attraction(int i)
{
  int x=ant[i].posx;
  int y=ant[i].posy;

  if (acces1(x-1,y)>acces1(x+1,y))
    ant[i].posx --;
  else
    if (acces1(x-1,y)<acces1(x+1,y))
      ant[i].posx ++;

  if (acces1(x,y-1)>acces1(x,y+1))
    ant[i].posy --;
  else
  if (acces1(x,y-1)<acces1(x,y+1))
    ant[i].posy ++;

}





/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


  int i
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
go_home(int i)
{
  int x=ant[i].posx;
  int y=ant[i].posy;


  if (nest_pheromon[y][x]<NEST_RADIUS)
	ant[i].state = searching;

  if (use_pheronom)
    draws(x,y,ant[i].food);

  if (nest_pheromon[y][x-1]>nest_pheromon[y][x+1])
    ant[i].posx ++;
  else
    if (nest_pheromon[y][x-1]<nest_pheromon[y][x+1])
      ant[i].posx --;

  if (nest_pheromon[y-1][x]>nest_pheromon[y+1][x])
    ant[i].posy ++;
  else
  if (nest_pheromon[y-1][x]<nest_pheromon[y+1][x])
    ant[i].posy --;

}





/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Check if the ant found food

int check_for_food 0 if food not found "food.rest" else

int i
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int
check_for_food(int i)
{
  int x=ant[i].posx;
  int y=ant[i].posy;

  for (i=0; i<nbfood; i++)
    {
      if (food[i].rest>0)
	if ((x<food[i].x+5) && (x>food[i].x-5) &&
	    (y<food[i].y+5) && (y>food[i].y-5))
	  {
	    food[i].rest -= ANT_STRENGTH;
	    if (food[i].rest <0)
	      food[i].rest=0;
	    return(food[i].rest);
	  }
    }
  return (0);
}







/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Trace mode for ants

  int i
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
trace_trail(int i)
{
  int x=ant[i].posx;
  int y=ant[i].posy;

  if (acces1(x,y)<active_trail_end)
	ant[i].state = searching;
  else
    {

      if ((acces1(x-1,y)>acces1(x+1,y)) &&
	  (nest_pheromon[y][x-1]>=nest_pheromon[y][x+1]))
	ant[i].posx --;
      else
	if ((acces1(x-1,y)<acces1(x+1,y)) &&
	    (nest_pheromon[y][x-1]<=nest_pheromon[y][x+1]))
	  ant[i].posx ++;

      if ((acces1(x,y-1)>acces1(x,y+1)) &&
	  (nest_pheromon[y-1][x]>=nest_pheromon[y+1][x]))
	ant[i].posy --;
      else
	if ((acces1(x,y-1)<acces1(x,y+1)) &&
	    (nest_pheromon[y-1][x]<=nest_pheromon[y+1][x]))
	  ant[i].posy ++;


      if ((x == ant[i].posx) && (y == ant[i].posy))
	{
	  if (nest_pheromon[y][x-1]>nest_pheromon[y][x+1])
	    ant[i].posx --;
	  else
	    if (nest_pheromon[y][x-1]<nest_pheromon[y][x+1])
	      ant[i].posx ++;

	  if (nest_pheromon[y-1][x]>nest_pheromon[y+1][x])
	    ant[i].posy --;
	  else
	    if (nest_pheromon[y-1][x]<nest_pheromon[y+1][x])
	      ant[i].posy ++;
	}
    }
}





/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Manage the ants behavior

  int i
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
ant_behavior(int i)
{
  int x=ant[i].posx;
  int y=ant[i].posy;
  int food;

  if (ant[i].state == searching)
    {
      if (acces1(x,y)>0)
	{
	if ((use_noise) && ((float)rand()/RAND_MAX < ant_noise_attraction))
	  random_walk(i);
	else
	  attraction(i);

	  if (acces1(x,y)>active_trail_start)
	    {
	      ant[i].state = trace;
	      nb_trace++;
	    }
	  else
	    nb_attr ++;
	}
      else
	{
	  random_walk(i);
	  nb_rand++;
	}
    }
  else
    if (ant[i].state == trace)
      {
	if ((use_noise) && ((float)rand()/RAND_MAX < ant_noise_trace))
	  random_walk(i);
	else
	  trace_trail(i);
	nb_trace ++;
      }
    else
      {
	if ((use_noise) && ((float)rand()/RAND_MAX < ant_noise_go_home))
	  random_walk(i);
	else
	  go_home (i);
	nb_go ++;
      }

  food = check_for_food(i);
  if (food != 0)
    {
      ant[i].state = going_home;
      ant[i].food = food;
    }
}
