/*   thread.c --- Ants Colony
 *
 *   Copyright (C) 2000 Valvassori Mose <djedi@altern.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA **/

#define THREAD_C


#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include "ants.h"
#include "../config.h"


#define TEST_TIME_LIMIT 150
#define TEST_FOOD_LIMIT 100

pthread_t working;
pthread_attr_t working_attr;
extern int s[];

int frame;							 /* Frame counter */
#ifdef STAT_DIR
FILE *randwm,*attrm,*tracem,*gohm,*foodm;
#endif
int nb_rand,nb_attr,nb_trace,nb_go,nb_food,step;
#ifdef HAVE_LIBJPEG
int save_jpeg=0;
char *save_jpeg_dir;
#endif

void stop_thread(void);

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  The running function of the thread.

void * thread

void *t
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void *
thread(void *t)
{
  int dummy;
  int i;
#ifdef STAT_DIR
  char monitor_name[255];
#else
#ifdef HAVE_LIBJPEG
  char monitor_name[255];
#endif
#endif
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,&dummy);

  init_world();
#ifdef STAT_DIR
  sprintf(monitor_name,"%s/%s",STAT_DIR,"random_walk");
  randwm=fopen(monitor_name,"w");
  sprintf(monitor_name,"%s/%s",STAT_DIR,"attraction");
  attrm=fopen(monitor_name,"w");
  sprintf(monitor_name,"%s/%s",STAT_DIR,"trace");
  tracem=fopen(monitor_name,"w");
  sprintf(monitor_name,"%s/%s",STAT_DIR,"home");
  gohm=fopen(monitor_name,"w");
  //  gohm=fopen("/tmp/home","w");
  sprintf(monitor_name,"%s/%d",STAT_DIR,nbants);
  do {
    foodm=fopen(monitor_name,"w");
  }while (foodm== NULL);
#endif
  step=0;

  for (i=0; i<nbants; i++)
    {
      init_ants(i);
    }

  frame=0;

  while (true)
    {
      /*  draw(MW/2+(rand()%16)-6,MH/2+(rand()%12)-6);
       *  draw(MW/4+(rand()%16)-6,MH/2+(rand()%12)-6);
       *  draw(MW/2+(rand()%16)-6,MH/4+(rand()%12)-6);
       *
       *  draw(MW/2+(rand()%16)-6,20+MH/2+(rand()%12)-6);
       *  draw(MW/2+(rand()%16)-6,40+MH/2+(rand()%12)-6);
       *  draw(MW/2+(rand()%16)-6,60+MH/2+(rand()%12)-6);
       *  draw(MW/2-25+(rand()%16)-6,60+MH/2+(rand()%12)-6);
       *  draw(MW/2-50+(rand()%16)-6,60+MH/2+(rand()%12)-6);
       *  draw(MW/2-75+(rand()%16)-6,60+MH/2+(rand()%12)-6); */

      for (i=0; i<nbants; i++)
	{
	  ant_behavior(i);
	}

      diffusion_evaporation_pheromon();
      world2image();
      XPutImage(display,sw->window,sw->gc,im,0,0,0,0,MW,MH);

#ifdef STAT_DIR
      fprintf(randwm,"%d\n",nb_rand);
      fprintf(attrm,"%d\n",nb_attr);
      fprintf(tracem,"%d\n",nb_trace);
      fprintf(gohm,"%d\n",nb_go);
      fprintf(foodm,"%d\n",nb_food);
#endif
      /*  if (nb_food != nbfood*DEFAULT_REST)
       *	step++;
       *
       *  if (nb_food == 0)
       *	break;
       *  if ((step > TEST_TIME_LIMIT) && (nb_food > (nbfood*DEFAULT_REST-TEST_FOOD_LIMIT)))
       *	{
       *	  printf("* missed *");
       *	  break;
       *	} */

      nb_rand=nb_attr=nb_trace=nb_go=nb_food=0;

#ifdef HAVE_LIBJPEG
      if (save_jpeg==1)
	{
	  sprintf(monitor_name,"%s/%05d.jpg",save_jpeg_dir,frame);
	  write_JPEG_file(monitor_name,90);
	}
#endif

      frame++;
    }

  /*  pthread_exit(NULL); */
}




void *
test(void *t)
{
  int i;

  for (nbants=nbants; nbants>0; nbants-- )/*  nbants/=1.1) */
    {
      do
	{
	  printf("- %d ",nbants); fflush(stdout);
	  thread(NULL);
#ifdef STAT_DIR
	  fclose(randwm);
	  fclose(attrm);
	  fclose(tracem);
	  fclose(gohm);
	  fclose(foodm);
#endif
	}
      while (step == TEST_TIME_LIMIT+1);
    }
}



/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  This function initialize abd start the thread
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
init_thread(void)
{
  int err;

  pthread_attr_init (&working_attr);
  if ((err = pthread_attr_setscope(&working_attr,PTHREAD_SCOPE_PROCESS)))
    /*  printf("%s:%d: warning: %s\n",__FILE__,__LINE__,strerror(err)) */; /* Should be check in 'Configure' */

  /*  pthread_create(&working,&working_attr,test,NULL);   */
  pthread_create(&working,&working_attr,thread,NULL);
}





/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  stop the thread
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
stop_thread(void)
{
#ifdef STAT_DIR
  fclose(randwm);
  fclose(attrm);
  fclose(tracem);
  fclose(gohm);
  fclose(foodm);
#endif
  pthread_cancel(working);
}
